module Parser where
import Prelude

import Control.Alt (class Alt, (<|>))
import Control.Alternative (class Alternative)
import Control.Lazy (class Lazy)
import Control.Plus (class Plus, empty)

import Data.Array as A
import Data.Int as Int
import Data.Maybe (Maybe(..))
import Data.String.CodeUnits (uncons, toCharArray, fromCharArray)
import Data.Tuple (Tuple(..))

import BaseFunctions (concat, isAlpha, isAlphaNum, isDigit, isLower, isSpace, isUpper)

newtype Parser v = Parser (String -> Maybe (Tuple v String))
parse :: forall a. Parser a -> (String -> Maybe (Tuple a String))
parse (Parser p) = p

item :: Parser Char
item = Parser p
  where p str = case uncons str of
                  Nothing -> Nothing
                  Just { head: x, tail: xs } -> Just (Tuple x xs)

instance functorParser :: Functor Parser where
  map f (Parser p) =
    Parser $ \s0 -> do
      (Tuple x s1) <- p s0
      pure (Tuple (f x) s1)

instance applyParser :: Apply Parser where
  apply (Parser p1) (Parser p2)
    = Parser $ \str -> case p1 str of
        Nothing -> Nothing
        Just (Tuple x1 xs1) -> case p2 xs1 of
          Nothing -> Nothing
          Just (Tuple x2 xs2) -> Just (Tuple (x1 x2) xs2)

instance applicativeParser :: Applicative Parser where
  pure a = Parser $ \str -> Just (Tuple a str)

instance bindParser :: Bind Parser where
  bind (Parser p) f = Parser $ \str -> case p str of
    Nothing -> Nothing
    Just (Tuple x xs) ->
      let (Parser p2) = f x
      in p2 xs

instance altParser :: Alt Parser where
  alt :: forall a. Parser a -> Parser a -> Parser a
  alt (Parser p1) (Parser p2) = Parser p
    where
      p stream = case p1 stream of
        Nothing -> p2 stream
        right -> right

instance plusParser :: Plus Parser where
  empty :: forall a. Parser a
  empty = Parser \_ -> Nothing

instance alternativeParser :: Alternative Parser

instance lazyParser :: Lazy (Parser a) where
  defer f = Parser \str -> parse (f unit) str

-- Generic parsing functions.

sat :: (Char -> Boolean) -> Parser Char
sat p = do x <- item
           if p x then pure x else empty

digit :: Parser Char
digit = sat isDigit

lower :: Parser Char
lower = sat isLower

upper :: Parser Char
upper = sat isUpper

letter :: Parser Char
letter = sat isAlpha

alphanum :: Parser Char
alphanum = sat isAlphaNum

char :: Char -> Parser Char
char x = sat (_ == x)

string :: String -> Parser String
string str = case A.uncons (toCharArray str) of
  Nothing -> Parser \stream -> Just (Tuple "" stream)
  Just { head: x, tail: xs } -> do c <- char x
                                   rest <- string (fromCharArray xs)
                                   pure (concat c rest)

ident :: Parser String
ident = do x  <- lower
           xs <- A.many alphanum
           pure (fromCharArray $ A.cons x xs)

nat :: Parser Int
nat = do xs <- A.some digit
         case Int.fromString (fromCharArray xs) of
           Nothing -> empty
           Just x -> pure x

int :: Parser Int
int = do _ <- char '-'
         n <- nat
         pure (-n)
       <|> nat

space :: Parser Unit
space = do _ <- A.many (sat isSpace)
           pure unit

token :: forall a. Parser a -> Parser a
token p = do space
             v <- p
             _ <- space
             pure v

identifier :: Parser String
identifier = token ident

natural :: Parser Int
natural = token nat

integer :: Parser Int
integer = token int

symbol :: String -> Parser String
symbol xs = token (string xs)

many1 :: forall a. Parser a -> Parser (Array a)
many1 p = do first <- p
             rest <- A.many p
             pure $ A.cons first rest
